#include	<stdio.h>

#include	"cmdopt.h"


//
// Internal functions.
//


// Move optind to tail, and slide other options.
static	void	CmdOpt_Slide( hCmdOpt h )
{
	int		i;
	char	*tmp;

	tmp = h->argv[ h->optind ];
	for ( i = h->optind; i < h->argc - 1; i++ )
		h->argv[i] = h->argv[ i + 1 ];
	h->argv[i] = tmp;

	h->nextchar = NULL;
	h->optnum--;
}

// Next argument.
static	void	CmdOpt_NextArg( hCmdOpt h )
{
	h->optind++;
	h->nextchar = NULL;
}

// Option (true) or not (false).
static	int		CmdOpt_IsOption( hCmdOpt h )
{
	int		ret = 1;
	const char	*arg = h->argv[ h->optind ];

	if ( *arg++ != '-' )
		return	0;

	if ( *arg == '-' )
		arg++, ret++;

	return	ret - ( *arg == '\0' );
}

// Get argument of option.
static	void	CmdOpt_GetArg( hCmdOpt h )
{
	// Next argument.
	if ( *h->nextchar == '\0' )
	{
		CmdOpt_NextArg( h );
		h->nextchar = h->argv[ h->optind ];
	}

	// Check whether the option has argument or not.
	if ( h->optind < h->optnum )
	{
		h->optarg = h->nextchar;
		CmdOpt_NextArg( h );
	}
	else
		h->optarg = NULL;
}

// Search option.
static	int		CmdOpt_Search( hCmdOpt h )
{
	const char	*ptr;

	// Option character.
	h->optopt = *h->nextchar++;

	for ( ptr = h->optstring; *ptr != '\0'; ptr++ )
	{
		if ( *ptr == h->optopt )
		{
			// Get argument, if required.
			if ( ptr[1] == ':' )
			{
				CmdOpt_GetArg( h );

				// No argument.
				if ( h->optarg == NULL && ptr[2] != ':' )
					return	':';
			}
			return	h->optopt;
		}
	}

	// End of options.
	if ( h->optopt == '-' )
	{
		CmdOpt_NextArg( h );
		while ( h->optind < h->optnum )
			CmdOpt_Slide( h );
		return	-1;
	}

	// Undefined option character.
	return	'?';
}

// Compare long option with argument, and return length.
static	int		CmdOpt_MatchLength( const char *opt, const char *arg )
{
	int		len = 0;

	// If invalid character is found, return 0.
	while ( *arg != '\0' && *arg != '=' )
	{
		if ( *arg++ != *opt++ )
			return	0;
		len++;
	}

	// Perfect match.
	if ( *arg == '\0' || *arg == '=' )
		return	-len;

	return	len;
}

// Check long option.
static	int		CmdOpt_MatchLong( hCmdOpt h )
{
	int		i, len;
	int		max = 0, max_optind = -1;

	// No long option.
	if ( h->longopts == NULL )
		return	max_optind;

	for ( i = 0; h->longopts[i].name != NULL; i++ )
	{
		len = CmdOpt_MatchLength( h->longopts[i].name, h->nextchar );
		if ( len < 0 )
		{
			// Perfect match.
			h->nextchar -= len;
			return	i;
		}
		else if ( len > max )
		{
			// Longest match.
			max = len;
			max_optind = i;
		}
		else if ( len == max )
			max_optind = -1;
	}

	h->nextchar += max;
	return	max_optind;
}

// Get argument of long option.
static	void	CmdOpt_GetLongArg( hCmdOpt h )
{
	if ( *h->nextchar == '=' )
	{
		h->optarg = h->nextchar + 1;
		CmdOpt_NextArg( h );
	}
	else
	{
		CmdOpt_NextArg( h );

		// Whether the long option has argument or not.
		if ( h->optind < h->optnum )
		{
			h->optarg = h->argv[ h->optind ];
			CmdOpt_NextArg( h );
		}
		else
			h->optarg = NULL;
	}
}

// Search long option.
static	int		CmdOpt_SearchLong( hCmdOpt h )
{
	const LongOpt	*option;

	// Long option.
	h->optlong = h->argv[ h->optind ];

	// Get option.
	if ( ( h->longindex = CmdOpt_MatchLong( h ) ) < 0 )
	{
		CmdOpt_NextArg( h );
		return	'?';
	}

	// Get argument, if required.
	option = h->longopts + h->longindex;
	if ( option->has_arg )
	{
		CmdOpt_GetLongArg( h );

		// No argument.
		if ( h->optarg == NULL )
			return	':';
	}
	else if ( *h->nextchar == '=' )
	{
		// Extra argument.
		CmdOpt_GetLongArg( h );
		return	'?';
	}

	// Return to other variable.
	if ( option->flag != NULL )
	{
		*option->flag = option->val;
		return	0;
	}

	return	option->val;
}

// Analyze command line options.
static	int		CmdOpt_Analyze( hCmdOpt h )
{
	int		type;

	// Initialize.
	h->optopt = 0;
	h->optlong = NULL;
	h->optarg = NULL;
	h->longindex = 0;

	while ( h->optind < h->optnum )
	{
		if ( h->nextchar == NULL )
		{
			// Check whether the argument is option or not.
			if ( !( type = CmdOpt_IsOption( h ) ) )
				CmdOpt_Slide( h );
			else
			{
				h->nextchar = h->argv[ h->optind ] + type;
				if ( type == 2 )
					return	CmdOpt_SearchLong( h );
			}
		}
		else
		{
			if ( *h->nextchar == '\0' )
			{
				// Next argument.
				CmdOpt_NextArg( h );
				continue;
			}

			// Search option character.
			return	CmdOpt_Search( h );
		}
	}

	return	-1;
}


//
// API.
//


// Get option from command line options.
int		CmdOpt_Get( hCmdOpt h )
{
	int		value;

	value = CmdOpt_Analyze( h );

	// Warning.
	if ( h->opterr )
	{
		if ( value == ':' )
		{
			// No argument.
			if ( h->optlong == NULL )
				printf( "option requires an argument -- %c\n", h->optopt );
			else
				printf( "option `--%s\' requires an argument\n",
						h->longopts[ h->longindex ].name );
		}
		else if ( value == '?' )
		{
			// Invalid option.
			if ( h->optlong == NULL )
				printf( "invalid option -- %c\n", h->optopt );
			else
				printf( "unrecognized option `%s\'\n", h->optlong );
		}
		else if ( value != -1 && h->opterr == 2 )
		{
			// Warning all.
			if ( h->optlong == NULL )
			{
				if ( h->optarg != NULL )
					printf( "option with `%s' -- %c\n", h->optarg, h->optopt );
				else
					printf( "option -- %c\n", h->optopt );
			}
			else
			{
				if ( h->optarg != NULL )
					printf( "option `--%s\' with `%s'\n",
							h->longopts[ h->longindex ].name, h->optarg );
				else
					printf( "option `--%s'\n",
							h->longopts[ h->longindex ].name );
			}
		}
	}

	return	value;
}

// Initialize structure.
void	CmdOpt_Init( hCmdOpt h, int argc, char **argv,
					 const char *optstring, const LongOpt *longopts )
{
	static	const char	optstring_empty[] = "";

	h->argc = argc;
	h->argv = argv;
	h->optnum = h->argc;

	h->longopts = longopts;
	h->optstring = ( optstring != NULL ) ? optstring : optstring_empty;

	h->optind = 1;
	h->nextchar = NULL;
	h->optarg = NULL;
	h->optopt = 0;
	h->optlong = NULL;
	h->opterr = 0;
	h->longindex = 0;
}
